'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var queueUtil = require('@nx-js/queue-util');
var React = require('react');
var React__default = _interopDefault(React);
var observerUtil = require('@nx-js/observer-util');
var PropTypes = _interopDefault(require('prop-types'));

// run state change reactions synchronously
// TODO: add batching in certain cases
const state = new queueUtil.Queue(queueUtil.priorities.SYNC);

// run integration reactions with a low priority
// URL and storage updates are not something the user is eagerly waiting for
const integrations = new queueUtil.Queue(queueUtil.priorities.LOW);

function stop() {
  state.stop();
  integrations.stop();
}

function start() {
  state.start();
  integrations.start();
}

function process() {
  state.process();
  integrations.process();
}

// this is used to save the component on the state for static lifecycle methods
const COMPONENT = Symbol('owner component');
// this used for empty setState calls to trigger component updates
const DUMMY_STATE = {};

function view(Comp, { devtool: rawDevtool } = {}) {
  const isStatelessComp = !(Comp.prototype && Comp.prototype.isReactComponent);
  const BaseComp = isStatelessComp ? React.Component : Comp;

  const devtool = rawDevtool ? operation => rawDevtool(Object.assign({ Component: Comp }, operation)) : undefined;

  // return a HOC which overwrites render, shouldComponentUpdate and componentWillUnmount
  // it decides when to run the new reactive methods and when to proxy to the original methods
  class ReactiveHOC extends BaseComp {
    constructor(props, context) {
      super(props, context);

      this.state = this.state || {};
      this.state[COMPONENT] = this;

      // create a reactive render for the component
      // run a dummy setState to schedule a new reactive render, avoid forceUpdate
      const updater = () => this.setState(DUMMY_STATE);
      // the used scheduler ca be stopped, restarted and flushed any time
      this.render = observerUtil.observe(this.render, {
        scheduler: {
          add: () => state.add(updater),
          delete: () => state.delete(updater)
        },
        debugger: devtool,
        lazy: true
      });
    }

    render() {
      return isStatelessComp ? Comp(this.props, this.context) : super.render();
    }

    // react should trigger updates on prop changes, while easyState handles store changes
    shouldComponentUpdate(nextProps, nextState) {
      const { props, state: state$$1 } = this;

      // respect the case when user prohibits updates
      if (super.shouldComponentUpdate && !super.shouldComponentUpdate(nextProps, nextState)) {
        devtool && devtool({ type: 'render', renderType: 'blocked' });
        return false;
      }

      // return true if it is a reactive render or state changes
      if (state$$1 !== nextState) {
        devtool && devtool({ type: 'render', renderType: 'reactive' });
        return true;
      }

      // the component should update if any of its props shallowly changed value
      const keys = Object.keys(props);
      const nextKeys = Object.keys(nextProps);
      if (nextKeys.length !== keys.length || nextKeys.some(key => props[key] !== nextProps[key])) {
        devtool && devtool({
          type: 'render',
          renderType: 'normal',
          props: nextProps,
          oldProps: props
        });
        return true;
      }
      return false;
    }

    // add a custom deriveStoresFromProps lifecyle method
    static getDerivedStateFromProps(props, state$$1) {
      if (super.deriveStoresFromProps) {
        // inject all local stores and let the user mutate them directly
        const stores = mapStateToStores(state$$1);
        super.deriveStoresFromProps(props, ...stores);
      }
      // respect user defined getDerivedStateFromProps
      if (super.getDerivedStateFromProps) {
        return super.getDerivedStateFromProps(props, state$$1);
      }
      return null;
    }

    componentWillUnmount() {
      // call user defined componentWillUnmount
      if (super.componentWillUnmount) {
        super.componentWillUnmount();
      }
      // clean up memory used by Easy State
      observerUtil.unobserve(this.render);
    }
  }

  ReactiveHOC.displayName = Comp.displayName || Comp.name;
  // static props are inherited by class components,
  // but have to be copied for function components
  if (isStatelessComp) {
    Object.assign(ReactiveHOC, Comp);
  }

  return ReactiveHOC;
}

function mapStateToStores(state$$1) {
  // find store properties and map them to their none observable raw value
  // to do not trigger none static this.setState calls
  // from the static getDerivedStateFromProps lifecycle method
  const component = state$$1[COMPONENT];
  return Object.keys(component).map(key => component[key]).filter(observerUtil.isObservable).map(observerUtil.raw);
}

const emptyTokens = new Set([undefined, '', '.']);
const notEmpty = token => !emptyTokens.has(token);

function normalizePath(path = '', depth) {
  const normalizedPath = [];
  for (let token of toPathArray(path)) {
    if (token === '..') {
      depth--;
    } else if (notEmpty(token)) {
      normalizedPath.push(token);
    }
  }

  return {
    normalizedPath,
    depth: path[0] === '/' ? 0 : depth
  };
}

// convert pathname strings to arrays
function toPathArray(path = '') {
  return path.split('/').filter(notEmpty);
}

// convert path arrays to absolute pathname strings
function toPathString(path = []) {
  // the leading '/' is important, it differentiates absolute pathnames from relative ones
  return '/' + path.filter(notEmpty).join('/');
}

// convert params objects to query strings
function toQuery(params = {}) {
  const tokens = [];

  for (let key in params) {
    let value = params[key];
    if (notEmpty(value)) {
      key = encodeURIComponent(key);
      // JSON stringify the value to keep the type information
      value = encodeURIComponent(JSON.stringify(value));
      tokens.push(`${key}=${value}`);
    }
  }
  return tokens.length ? `?${tokens.join('&')}` : '';
}

// converts &key=value strings to objects
function toParams(queryString = '') {
  const queryTokens = queryString
  // remove the leading ? or # character
  .slice(1).split('&').filter(notEmpty);

  const params = {};
  for (let token of queryTokens) {
    const keyValue = token.split('=');
    const key = decodeURIComponent(keyValue[0]);
    // JSON.stringify and parse keeps the type information
    const value = JSON.parse(decodeURIComponent(keyValue[1]));
    params[key] = value;
  }
  return params;
}

// convert scroll anchor to URL hash
function toHash(scroll = {}) {
  return scroll.anchor ? `#${scroll.anchor}` : '';
}

// convert url hash to scroll anchor
function toScroll(hash) {
  return hash ? { anchor: hash.slice(1) } : undefined;
}

// convert object to url
function toUrl({ path, params, scroll }) {
  return toPathString(path) + toQuery(params) + toHash(scroll);
}

// convert url to object
function toObject(url) {
  let queryIndex = url.indexOf('?');
  queryIndex = queryIndex === -1 ? Infinity : queryIndex;
  let hashIndex = url.indexOf('#');
  hashIndex = hashIndex === -1 ? Infinity : hashIndex;

  return {
    path: toPathArray(url.slice(0, queryIndex)),
    params: toParams(url.slice(queryIndex, hashIndex)),
    scroll: toScroll(url.slice(hashIndex))
  };
}

// augments the props with extraProps, which are not in excludeProps
// nice for proxying irrelevant props to the underlying DOM element
function addExtraProps(props, extraProps, excludeProps) {
  for (let key in extraProps) {
    if (!(key in excludeProps)) {
      props[key] = extraProps[key];
    }
  }
  return props;
}

const routers = [];
let routingStatus;
const initStatuses = new Set();

// register a router at a given depth
function registerRouter(router, depth) {
  // there can be multiple parallel routers at a given depth (we need a Set)
  if (!routers[depth]) {
    routers[depth] = new Set();
  }
  routers[depth].add(router);
  // route the newly added router, if we are not routing currently
  // or if the routing process is already past the current depth
  // this may happen if the router is not added because of the routing process
  // but was hidden some by conditional or lazy loading for example
  if (!routingStatus || depth <= routingStatus.depth) {
    initRouter(router, depth);
  }
}

// router newly added routers even if there is no ongoing routing process
function initRouter(router, depth) {
  const status = { depth, cancelled: false };
  initStatuses.add(status);

  return Promise.resolve().then(() => router.startRouting()).then(
  // cancel the routing if a new routing started in the meantime
  resolvedData => !status.cancelled && router.finishRouting(resolvedData)).then(() => initStatuses.delete(status));
}

// release routers on componentWillUnmount
function releaseRouter(router, depth) {
  routers[depth].delete(router);
}

// this is part of the public API, it triggers root level routings
function route(args) {
  routeFromDepth(args, 0);
}

// this cancels ongoing routings and recursively routes all routers
function routeFromDepth({
  to = toPathString(path),
  params: toParams$$1 = {},
  scroll,
  push,
  inherit
} = {}, depth = 0) {
  const { depth: normalizedDepth, normalizedPath } = normalizePath(to, depth);
  depth = normalizedDepth;

  // there may be routers which route outside of the standard routing process
  // for example a router was added by lazy loading and does its initial routing
  // cancel routings for these routers
  initStatuses.forEach(status => status.cancelled = depth <= status.depth);
  // cancel the standard ongoing routing process if there is one
  if (routingStatus) {
    routingStatus.cancelled = true;
  } else {
    // flush the pending low priority URL changes to have a consistent state before starting a new routing
    // only flush it if there is no ongoing routing
    // otherwise the old and new routing should be treated in one batch to avoid flicker
    integrations.process();
  }
  // stop all schedulers until the end of the routing and commit them at once
  // this includes state based view updates, URL updates and localStorag updats
  // having view updates during the routing can cause flickers
  // because of the random onRoute data resolve timings, it is nicer to commit everything at once
  stop();

  // create a new routing status
  // this may be cancelled by future routing processes
  const status = routingStatus = { depth, cancelled: false };

  // update the path array with the desired new path
  // and save the old path for comparison at the end of the routing
  const prevPath = Array.from(path);
  path.splice(depth, Infinity, ...normalizedPath);
  // replace or extend the query params with the new params
  // only mutate the params object, never replace it (because it is an observable)
  if (!inherit) {
    Object.keys(params).forEach(key => delete params[key]);
  }
  Object.assign(params, toParams$$1);

  // recursively route all routers, then finish the routing
  return Promise.resolve().then(() => switchRoutersFromDepth(depth, status)).then(() => finishRouting({ scroll, push, prevPath }, status));
}

// this recursively routes all parallel routers form a given depth
// all routers are finished routing at a depth before continuing with the next depth
function switchRoutersFromDepth(depth, status) {
  const routersAtDepth = Array.from(routers[depth] || []);

  if (routersAtDepth.length) {
    return Promise.resolve()
    // during the first routing half, the user may intercept the routing
    // and cancel it by triggering a new routing
    // we have to wait for all routers before we can go on
    // to make sure the routing was not cancelled
    .then(() => startRoutingAtDepth(routersAtDepth, status))
    // after all routers finished the first half, all of them does the second half
    .then(resolvedData => finishRoutingAtDepth(routersAtDepth, resolvedData, status))
    // all routers finished routing at this depth, go on with the next depth
    .then(() => switchRoutersFromDepth(++depth, status));
  }
}

// do the first routing half if the routing is not cancelled
// it includes data resolving, interception and lazy loading
function startRoutingAtDepth(routersAtDepth, status) {
  if (!status.cancelled) {
    return Promise.all(routersAtDepth.map(router => router.startRouting()));
  }
}

// do the second routing half if the routing is not cancelled
// it includes rendering and animation
function finishRoutingAtDepth(routersAtDepth, resolvedData, status) {
  if (!status.cancelled) {
    return Promise.all(routersAtDepth.map((router, i) => router.finishRouting(resolvedData[i])));
  }
}

// all routers updated recursively by now, it is time to finish the routing
// if it was not cancelled in the meantime
function finishRouting({ scroll, push, prevPath }, status) {
  if (!status.cancelled) {
    const pathChanged = toPathString(path) !== toPathString(prevPath);
    // push a new history item when necessary
    // it is important to do this before restarting the schedulers
    // to apply all new history replace operations to the new item
    if (push === true || push !== false && pathChanged) {
      history.push({ path, params, scroll });
    }

    // flush the URL updates in one batch and restart the automatic processing
    // it is important to call this after handleHistory()
    // in case of a newly pushed history item, the flushed URL changes
    // should replace the new item instead of the old one
    // also flush the view updates
    process();
    start();
    // the routing is over and the is no currently ongoing routing process
    routingStatus = undefined;
  }
}

const elements = {
  anchor: 'a',
  div: 'div'
};

const scroller = {
  scrollTo() {}
};

const animation = {
  setup() {},
  enter() {},
  leave() {}
};

const storage = observerUtil.observable({});
const params = observerUtil.observable({});
const path = observerUtil.observable([]);

// TODO: maybe this needs an emtpy object inside!
const items = [];
let idx = 0;
const history = {
  createItem(item = {}) {
    if (typeof item === 'string') {
      item = toObject(item);
    }
    item = {
      path: Array.from(item.path),
      params: Object.assign({}, item.params),
      // scroll config can be passed without copying as it is read-only
      scroll: item.scroll,
      url: toUrl(item)
    };
    // update the params and path to reflect the new history item
    Object.assign(params, item.params);
    path.splice(0, Infinity, ...item.path);
    return item;
  },
  get state() {
    return items[idx];
  },
  get length() {
    return items.length;
  },
  push(item) {
    items.splice(++idx, Infinity, this.createItem(item));
  },
  replace(item) {
    items[idx] = this.createItem(item);
  },
  go(toIdx) {
    idx = Math.min(items.length - 1, Math.max(0, idx + toIdx));
    const { path, params, scroll } = items[idx];
    return route({
      to: toPathString(path),
      params,
      scroll,
      push: false
    });
  },
  back() {
    this.go(-1);
  },
  forward() {
    this.go(1);
  }
};

function syncHistory() {
  const scroll = history.state ? history.state.scroll : undefined;
  history.replace({ path, params, scroll });
}
observerUtil.observe(syncHistory, { scheduler: integrations });

// Router selects a single child to render based on its children's page props
// and the URL pathname token at the Router's depth (they can be nested)
class Router extends React.PureComponent {
  constructor(...args) {
    var _temp;

    return _temp = super(...args), this.state = {}, this.saveContainer = container => this.container = container, _temp;
  }

  getChildContext() {
    return { easyRouterDepth: this.depth + 1 };
  }

  // depth stores how nested is the router, root routers have a depth of 0
  get depth() {
    return this.context.easyRouterDepth || 0;
  }

  componentDidMount() {
    registerRouter(this, this.depth);
  }

  componentWillUnmount() {
    releaseRouter(this, this.depth);
  }

  // the raw DOM container node is needed for the animations


  // this is part of the public API
  // it routes every router from this depth (including this one)
  route(routingOptions) {
    routeFromDepth(routingOptions, this.depth);
  }

  // routing is split in 2 phases
  // first all parallel routers at the same depth executes startRouting
  // then all parallel routers at the same depth execute finishRouting
  startRouting() {
    const { onRoute } = this.props;
    const fromPage = this.state.page;
    const toPage = path[this.depth] || this.props.defaultPage;

    // fill the path with the default page, if the current path token is empty
    // this is important for relative links and automatic active link highlight
    path[this.depth] = toPage;

    // onRoute is where do user can intercept the routing or resolve data
    if (onRoute) {
      return onRoute({
        target: this,
        fromPage,
        toPage,
        fromParams: history.state.params
      });
    }
  }

  // finishRouting is called when all parallel routers at the current depth
  // finished executing startRouting
  // resolvedData is the data returned from props.onRoute in startRouting
  finishRouting(resolvedData) {
    const fromPage = this.state.page;
    const toPage = path[this.depth];
    const { enterAnimation, leaveAnimation } = this.props;
    let shouldAnimate = this.props.shouldAnimate;

    if (typeof shouldAnimate === 'function') {
      shouldAnimate = shouldAnimate({ fromPage, toPage });
    }
    if (shouldAnimate === undefined) {
      shouldAnimate = fromPage !== toPage;
    }
    // this typically saves the current view to use later for cross fade effects
    // the current view is soon replaced by setState, so this is necessary
    if (shouldAnimate && this.container) {
      animation.setup(this.container);
    }

    const nextState = {
      resolvedData,
      page: toPage
    };

    // render the new page with the resolvedData
    return new Promise(resolve => this.setState(nextState, resolve)).then(() => {
      // run the animations when the new page is fully rendered, but do not wait for them
      // the views may be hidden by the animation, but the DOM routing is already over
      // it is safe to go on with routing the next level of routers
      if (shouldAnimate && this.container) {
        // only do an enter animation if this is not the initial routing of the router
        // this prevents cascading over-animation, in case of nested routers
        // only the outmost one will animate, the rest will appear normally
        if (enterAnimation && this.inited) {
          animation.enter(this.container, enterAnimation);
        }
        // leave must come after render
        // it is re-appending the old dom to the container
        // doing this before enter will confuse the render about which DOM
        // to animate and it will try to animate the old one twice instead of both once
        if (leaveAnimation) {
          animation.leave(this.container, leaveAnimation);
        }
      }
      // the router has done at least one full routing
      this.inited = true;
    });
  }

  render() {
    const { element } = this.props;
    const { page, resolvedData } = this.state;

    // render nothing if no matching view is found
    let toChild = null;
    // if the resolvedData from onRoute is a React element use it as the view
    // this allows lazy loading components (and virtual routing)
    if (React__default.isValidElement(resolvedData)) {
      // validate the resolved child to have a page prop
      validateChild(resolvedData);
      toChild = resolvedData;
    } else {
      // select the next child based on the children's page prop
      // and the string token in the URL pathname at the routers depth
      toChild = this.selectChild(page);
      if (toChild && resolvedData) {
        // of there is resolvedData from onRoute, inject it as props to the next view
        toChild = React__default.cloneElement(toChild, resolvedData);
      }
    }

    return React__default.createElement(element,
    // forward none Router specific props to the underlying DOM element
    addExtraProps({ ref: this.saveContainer }, this.props, Router.propTypes),
    // render the selected child as the only child element
    toChild);
  }

  // select the next view based on the children's page prop
  // and the string token in the URL pathname at the routers depth
  selectChild(page) {
    const children = React.Children.toArray(this.props.children);
    // validate if all children has a page prop
    children.forEach(validateChild);
    const selectedChild = children.find(child => child.props.page === page);
    // if the router is mounted and it has no matching child view,
    // try to render a notFoundPage
    // if the router is not yet mounted, the initial routing is not yet finished
    // in this case it should render nothing (null), instead of a notFoundPage
    if (!selectedChild && this.container) {
      const { notFoundPage } = this.props;
      return children.find(child => child.props.page === notFoundPage);
    }
    return selectedChild;
  }
}

// all Router children must have a page prop
Router.propTypes = {
  defaultPage: PropTypes.string.isRequired,
  notFoundPage: PropTypes.string,
  onRoute: PropTypes.func,
  enterAnimation: PropTypes.func,
  leaveAnimation: PropTypes.func,
  shouldAnimate: PropTypes.oneOfType([PropTypes.bool, PropTypes.func]),
  element: PropTypes.oneOfType([PropTypes.string, PropTypes.func])
};
Router.defaultProps = {
  element: elements.div
};
Router.childContextTypes = { easyRouterDepth: PropTypes.number };
Router.contextTypes = { easyRouterDepth: PropTypes.number };
function validateChild(child) {
  if (typeof child.props.page !== 'string') {
    throw new Error('Every Router child must have a string valued page prop');
  }
}

// Link is used to navigate between pages
// it can be relative ('home') or absolute ('/home'), just like vanilla HTML links
class Link extends React.PureComponent {
  constructor(...args) {
    var _temp;

    return _temp = super(...args), this.state = {}, this.onClick = ev => {
      const { to, params: params$$1, scroll, push, inherit, onClick } = this.props;
      // route all Routers from below the links depth
      // absolute links have a depth of 0
      routeFromDepth({ to, params: params$$1, scroll, push, inherit }, this.depth);
      // prevent the default behavior of anchor clicks (page reload)
      ev.preventDefault();
      // respect user defined onClick handlers on the Link
      if (onClick) {
        onClick(ev);
      }
    }, _temp;
  }

  // absolute links ('/home') have a depth of 0
  // relative links have the depth of their closest router ancestor
  get depth() {
    const { to } = this.props;
    const depth = this.context.easyRouterDepth || 0;
    const isRelative = !to || to[0] !== '/';
    return isRelative ? depth : 0;
  }

  // gets the full path for relative and absolute links too
  get absolutePath() {
    const { depth, normalizedPath } = normalizePath(this.props.to, this.depth);
    return path.slice(0, depth).concat(normalizedPath);
  }

  // automatically update the link activity on pathname and params changes
  // with a low priority scheduler
  componentDidMount() {
    this.activityUpdater = observerUtil.observe(() => this.setState({ isActive: this.isLinkActive() }), { scheduler: state });
  }

  // clean up transparent reactivity connections
  componentWillUnmount() {
    observerUtil.unobserve(this.activityUpdater);
  }

  isLinkActive() {
    const { activeClass, activeStyle, isActive, params: params$$1, scroll } = this.props;
    // only calculate link activity if there is an activeClass or activeSyle prop
    // otherwise it is not needed
    if (activeClass || activeStyle) {
      // let the user fine tune link activity with an isActive function prop
      if (isActive) {
        return isActive({ path: this.absolutePath, params: params$$1, scroll });
      }
      return this.isLinkPathActive() && this.isLinkParamsActive() && this.isLinkScrollActive();
    }
  }

  isLinkPathActive() {
    // URL pathname tokens before this.depth always match with the link
    // otherwise the link and its containing Router would not be rendered
    // URL pathname tokens after the link does not affect the check
    // for example '/profile' link matches with '/profile/settings' URL
    return this.absolutePath.every((page, i) => page === path[i]);
  }

  isLinkParamsActive() {
    const linkParams = this.props.params;
    if (linkParams) {
      // extra query params does not affect the match
      // for example { a: 1 } link params matches with { a: 1, b: 2 } URL query
      const paramKeys = Object.keys(linkParams);
      return paramKeys.every(key => linkParams[key] === params[key]);
    }
    return true;
  }

  isLinkScrollActive() {
    const linkScroll = this.props.scroll;
    const historyScroll = history.state.scroll;
    // scroll positions match when the link or the URL is missing a scroll anchor
    // or when the two anchors macth
    if (linkScroll && historyScroll) {
      return linkScroll.anchor === historyScroll.anchor;
    }
    return true;
  }

  render() {
    let {
      to,
      params: params$$1,
      scroll,
      element,
      children,
      activeClass,
      activeStyle,
      style,
      className
    } = this.props;
    const { isActive } = this.state;
    const { onClick } = this;

    // calculate a full link href for correct 'Open link in new tab' behavior
    const href = toUrl({
      path: this.absolutePath,
      params: params$$1,
      scroll
    });

    if (isActive) {
      // both are empty strings by default
      className = `${className} ${activeClass}`;
      // both are empty objects by default
      style = Object.assign({}, style, activeStyle);
    }

    return React__default.createElement(element,
    // forward none Link specific props to the rendered element
    addExtraProps({ href, onClick, className, style }, this.props, Link.propTypes), children);
  }
}
Link.propTypes = {
  to: PropTypes.string,
  params: PropTypes.object,
  scroll: PropTypes.object,
  push: PropTypes.bool,
  inherit: PropTypes.bool,
  onClick: PropTypes.func,
  className: PropTypes.string,
  style: PropTypes.object,
  activeClass: PropTypes.string,
  activeStyle: PropTypes.object,
  isActive: PropTypes.func,
  element: PropTypes.oneOfType([PropTypes.string, PropTypes.func])
};
Link.defaultProps = {
  element: elements.anchor,
  className: '',
  activeClass: '',
  style: {},
  activeStyle: {}
};
Link.contextTypes = {
  easyRouterDepth: PropTypes.number
};

exports.store = observerUtil.observable;
exports.view = view;
exports.Router = Router;
exports.Link = Link;
exports.route = route;
exports.elements = elements;
exports.scroller = scroller;
exports.animation = animation;
exports.storage = storage;
exports.params = params;
exports.path = path;
exports.history = history;
